# typed: false
# frozen_string_literal: true

module Setsuzoku
  # Core service required for a web service plugin.
  # It should be able to register all of its available api and auth strategies.
  # It acts as the orchestration between a plugin, auth_strategy, and api_strategy.
  module Service
    autoload :WebService, 'setsuzoku/service/web_service'

    extend Forwardable
    extend T::Sig
    extend T::Helpers
    abstract!

    attr_accessor :plugin
    attr_accessor :auth_strategy
    attr_accessor :api_strategy
    attr_accessor :external_api_handler
    def_delegators :@auth_strategy, :new_credential!, :set_credential!
    def_delegators :@api_strategy, :call_external_api, :request_class

    def self.included(klass)
      klass.extend(ClassMethods)
    end

    module ClassMethods
      extend T::Sig
      extend T::Helpers
      abstract!

      # The api and auth strategies available for the service.
      # api: { simple_name: ApiModule }
      #
      # @return [Hash(Hash(Class))] the available_strategies object for the Service.
      sig { abstract.returns(T::Hash[Symbol, T::Hash[Symbol, Class]]) }
      def available_strategies; end
    end

    # Initialize the plugin_service by setting its auth and api strategies.
    #
    # @param auth_strategy [AuthStrategy] the auth_strategy for the service to use.
    # @param api_strategy [ApiStrategy] the api_strategy for the service to use.
    #
    # @return [Service] the new instance of service with its correct strategies.
    sig(:final) do
      params(
          plugin: T.untyped,
          strategies: T::Hash[Symbol, T.untyped],
          args: T.untyped
      ).returns(T.any(Setsuzoku::Service::WebService::Service, T.untyped))
    end
    def initialize(plugin:, strategies:, **args)
      #TODO: here we need to assign credentials etc, I think.
      self.plugin = plugin
      self.external_api_handler = Setsuzoku.external_api_handler.new

      # iterate over all strategies this plugin's service uses and set their configuration
      strategies.each do |strategy, attrs|
        # get the strategy type from auth_strategy/api_strategy etc.
        type = T.must(strategy.to_s.split("_strategy").first).to_sym
        strat = attrs[:strategy]
        # associate the strategy with the service
        self.send("#{strategy}=", self.class.available_strategies[type][strat].new(service: self, credential: args[:credential], **attrs.except(:strategy)))
      end

      self
    end
  end
end